#!/usr/bin/python3

import argparse
from enum import Enum
import networkx as nx
import os
import re
import sys


class IBGUID:
    def __init__(self, guid: str):
        guid = guid.strip().replace('0x', '').replace('0X', '')
        if not re.fullmatch('[0-9a-fA-F]{1,16}', guid):
            raise ValueError('guid should be 1-16 hexadecimal numbers')
        # Some guids have leading '0' stripped
        guid = '0' * (16 - len(guid)) + guid
        self.guid = bytearray.fromhex(guid)

    def __str__(self):
        return self.guid.hex()

    def __hash__(self):
        return hash(self.guid.hex())

    def __eq__(self, other):
        return self.guid == other.guid


class IBNode:
    class Type(Enum):
        SWITCH = 1
        CA = 2

    @classmethod
    def from_nodeid(cls, nodeid):
        m = re.fullmatch('(?P<type>[SH])-(?P<guid>[0-9a-f]{16})',
                         nodeid
                        )
        if not m:
            raise RuntimeError("Can't parse nodeid " + nodeid)
        node = IBNode(m.group('guid'))
        if m.group('type') == 'S':
            node.type = IBNode.Type.SWITCH
        elif m.group('type') == 'H':
            node.type = IBNode.Type.CA
        node.nodeid = nodeid
        return node

    def __init__(self, guid: str = None):
       if guid:
            self.guid = IBGUID(guid)
       else:
            self.guid = None

    def __hash__(self):
        if not self.guid:
            raise ValueError('Can\'t hash IBNode without guid')
        return hash(self.guid)

    def __eq__(self, other):
        return self.guid == other.guid

    def __bool__(self):
        return self.guid is not None

    def __str__(self):
        if not self.guid:
            return '<Node without guid>'
        if self.type == IBNode.Type.CA:
            return "CA " + str(self.guid)
        elif self.type == IBNode.Type.SWITCH:
            return "Swtich " + str(self.guid)
        else:
            return "Unknow node " + str(self.guid)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        if key not in self.__dict__:
            self.__dict__[key] = value
            self.try_update_nodeid()
        elif self.__dict__[key] != value:
            raise ValueError('Trying to update with conflicting info: ' + str(value) + '-> ' + str(self) + '[' + key + '] = ' + str(value))

    def set_guid(self, guid):
        if self.guid:
            raise ValueError('Attempt to redefine IBNode guid')
        if isinstance(guid, IBGUID):
            self.guid = guid
        else:
            self.guid = IBGUID(guid)
        self.try_update_nodeid()

    def try_update_nodeid(self):
        if hasattr(self, 'nodeid'):
            return
        if hasattr(self, 'type') and self.guid:
            if self.type == IBNode.Type.SWITCH:
                self.nodeid = "S-" + str(self.guid)
            elif self.type == IBNode.Type.CA:
                self.nodeid = "H-" + str(self.guid)
            print("Set nodeid to", self.nodeid)

    def update(self, other_node):
        """
        Update self with new data from other_node
        If node and other_node have keys with different values,
        raise an exception
        """
        for p in other_node.__dict__:
            if p in self.__dict__:
                if self.__dict__[p] != other_node.__dict__[p]:
                    raise ValueError('Trying to update with conflicting info: ' + str(other_node.__dict__[p]) + '-> ' + str(self) + '[' + p + '] = ' + str(self.__dict__[p]))
            self.__dict__[p] = other_node.__dict__[p]


class IBNetwork:
    def __init__(self):
        self.nodes = {}
        self.connections = {}

    def add_node(self, node):
        """
        Add a node to network
        If node with the same guid exists, add new information.
        If old node and new node have the same member (e.g. vendid)
        with different values, raise exception
        """
        if not node:
            raise ValueError("Can't add node without guid")
        old_node = self.find_node_byguid(node.guid)
        if not old_node:
            self.nodes[node.guid] = node
        else:
            old_node.update(node)

    def add_connection(self, node1, node2):
            node1_known = self.find_node_byguid(node1.guid)
            node2_known = self.find_node_byguid(node2.guid)
            if not node1_known or not node2_known:
                raise RuntimeError("Can't find node " + str(node1) + " or " + str(node2))
            if node1_known not in self.connections:
                self.connections[node1_known] = {}
            if node2_known not in self.connections:
                self.connections[node2_known] = {}
            self.connections[node1_known][node2_known] = True
            self.connections[node2_known][node1_known] = True

    def check_connection(self, node1, node2):
        if node1 in self.connections:
            return node2 in self.connections[node1]

    def connections_by_node(self, node):
        if node in self.connections:
            return list(self.connections[node])
        else:
            return []

    def find_node_byguid(self, guid: IBGUID):
        if guid in self.nodes:
            return self.nodes[guid]
        else:
            return None

    def find_node_bynodeid(self, nodeid):
        new_node = IBNode.from_nodeid(nodeid)
        old_node = self.find_node_byguid(new_node.guid)
        return old_node or None

parser = argparse.ArgumentParser()
parser.add_argument('ibnetdiscover_output',
                    type=argparse.FileType('r'),
                    metavar='<ibnetdiscover output>',
                   )
parser.add_argument('-f', '--full-scope',
                    type=argparse.FileType('w'),
                    required=False,
                    help='Output scopefile. If not given, use stdout.',
                    default=sys.stdout,
                   )
# Distances are not used now, so calculate them only on demand
parser.add_argument('-d', '--distances',
                    type=argparse.FileType('w'),
                    required=False,
                    help='File to write distances between all TOR pairs. If omitted, distances will not be calculated.',
                   )
args = parser.parse_args()

print("Started", file=sys.stderr)

network = IBNetwork()

cur_node = IBNode()
for line in args.ibnetdiscover_output:
    line = line.strip()
    if line == '':
        if cur_node:
            network.add_node(cur_node)
            #print("Added node " + str(cur_node))
        cur_node = IBNode()
    elif line[0] == '[':
        #  [1](b8599f03004af86d) 	"S-0c42a1030060ac30"[2]		# lid 230 lmc 0 "MF0;rvn-everest-t1:MQM8700/U1" lid 2 4xHDR
        #  [19]	"S-043f720300ea0662"[11]		# "MF0;rvn-vega-l1:MQM8700/U1" lid 6 4xHDR
        #  [41]	"H-1c34da0300728cce"[1](1c34da0300728cce) 		# "Mellanox Technologies Aggregation Node" lid 75 4xHDR
        m = re.fullmatch('\[(?P<srcport>\d+)\](\((?P<srcportguid>[0-9a-f]+)\))?\s*"(?P<dstnodeid>[^"]+)"\[(?P<dstport>\d+)\](\((?P<dstportguid>[0-9a-f]+)\))?\s*#.*"(?P<dstnodedesc>[^"]+)".*',
                            line,
                            re.IGNORECASE | re.ASCII
                        )
        if not m:
            raise RuntimeError("Can't parse line " + line)
        dstnodeid = m.group('dstnodeid')
        dst_node = network.find_node_bynodeid(dstnodeid)
        if not dst_node:
            dst_node = IBNode.from_nodeid(dstnodeid)
            network.add_node(dst_node)
        network.add_connection(cur_node, dst_node)

    elif line.startswith("Switch"):
        #  Switch	41 "S-1c34da0300728cc6"		# "MF0;rvn-vega-t3:MQM8700/U1" enhanced port 0 lid 74 lmc 0
        m = re.fullmatch('Switch\s+(?P<numports>\d+)\s+"(?P<nodeid>[^"]+)"\s*#\s*"(?P<nodedesc>[^"]+)"\s*(?P<comment>.*)',
                            line,
                            re.IGNORECASE | re.ASCII
                        )
        if not m:
            raise RuntimeError("Can't parse line " + line)
        cur_node.numports = int(m.group('numports'))
        cur_node.nodeid = m.group('nodeid')
        cur_node.nodedesc = m.group('nodedesc')
        cur_node.comment = m.group('comment')
        network.add_node(cur_node)
    elif line.startswith("Ca"):
        #  Ca	1 "H-98039b0300860c37"		# "clx-everest-003 HCA-3"
        m = re.fullmatch('Ca\s+(?P<numports>\d+)\s+"(?P<nodeid>[^"]+)"\s*#\s*"(?P<nodedesc>[^"]+)"',
                            line,
                            re.IGNORECASE | re.ASCII
                        )
        if not m:
            raise RuntimeError("Can't parse line " + line)
        cur_node.numports = int(m.group('numports'))
        cur_node.nodeid = m.group('nodeid')
        cur_node.nodedesc = m.group('nodedesc')
        network.add_node(cur_node)
    elif re.fullmatch('(?P<propname>[a-z]+)=(?P<propval>0x[0-9a-f]+)(\((?P<portguid>[0-9a-f]+)\))?',
                        line,
                        re.IGNORECASE | re.ASCII
                        ):
        #  vendid=0x2c9
        #  devid=0x101b
        #  sysimgguid=0xb8599f03004af86c
        #  caguid=0xb8599f03004af86d
        #  switchguid=0xb8599f0300557370(b8599f0300557370)
        m = re.fullmatch('(?P<propname>[a-z]+)=(?P<propval>0x[0-9a-f]+)(\((?P<portguid>[0-9a-f]+)\))?',
                            line,
                            re.IGNORECASE | re.ASCII
                        )
        propname = m.group('propname')
        propval = m.group('propval')
        if propname == 'vendid' or propname == 'devid':
            cur_node[propname] = propval
        elif propname == 'sysimgguid':
            cur_node.sysimgguid = IBGUID(propval)
        elif propname == 'caguid':
            cur_node.set_guid(propval)
            cur_node.type = IBNode.Type.CA
        elif propname == 'switchguid':
            cur_node.set_guid(propval)
            cur_node.type = IBNode.Type.SWITCH
            # Ignore portguid for now
            # If we have connections to the same node to different ports
            # a conflict can occur
            #cur_node.portguid = IBGUID(m.group('portguid'))
        else:
            raise RuntimeError('Unknow property ' + propname + '=' + propval)

print("Parsing ibnetdiscover output file finished", file=sys.stderr)

MIN_NODES_IN_TOR = 4
tors = []
tor_nodes = {}

is_node = lambda n: n.type == IBNode.Type.CA
is_switch = lambda n: n.type == IBNode.Type.SWITCH

for n in filter(is_switch, network.nodes.values()):
    connected_nodes = list(filter(is_node, network.connections_by_node(n)))
    n.num_connected_nodes = len(connected_nodes)
    if n.num_connected_nodes >= MIN_NODES_IN_TOR:
        tors.append(n)
        tor_nodes[n] = connected_nodes

for tor_num, tor in enumerate(tors):
    for node in tor_nodes[tor]:
        args.full_scope.write("{},{},{}\n".format(node.nodedesc, tor.nodeid, tor_num))
if args.full_scope != sys.stdout:
    args.full_scope.close()

if not args.distances:
    print("Finished, no distances calculated", file=sys.stderr)
    sys.exit(0)

# Graph of switches only
g = nx.Graph()

for s in filter(is_switch, network.nodes.values()):
    g.add_node(s)
    for d in filter(is_switch, network.connections_by_node(s)):
        g.add_edge(s, d)

print("TORs: " + str(len(tors)) + " found", file=sys.stderr)

print("Calcucating distances between TORs, it may take significant time (up to tens of minutes for large networks)", file=sys.stderr)

lengths = {}
for s in tors:
    lengths[s] = nx.single_source_shortest_path_length(g, s)

for s in tors:
    for d in tors:
        if s != d:
            print("{} -- {} == {}".format(s, d, lengths[s][d]), file= args.distances)
if args.distances and args.distances != sys.stdout:
    args.distances.close()

print("Finished.", file=sys.stderr)
